package de.hdm_stuttgart.mi.sd1.task2;

import de.hdm_stuttgart.mi.exam.unitmarking.ExaminationTestDefaults;
import de.hdm_stuttgart.mi.exam.unitmarking.Marking;
import de.hdm_stuttgart.mi.sd1.ignore_me.ObjectWrapper;
import org.junit.Assert;
import org.junit.FixMethodOrder;
import org.junit.Test;
import org.junit.runners.MethodSorters;
@FixMethodOrder(MethodSorters.NAME_ASCENDING)

public class TestSquarePolynom extends ExaminationTestDefaults {

    @Test
    @Marking(points = 5)
    public void test_100_constructor_getValue() {

        assertPolynomValue(0, 0, 3, 20.5);
        assertPolynomValue(0, 0, 3, -10.);
        assertPolynomValue(0, 0, 3, 75.);

        assertPolynomValue(-1, 3, 5, 2.);
        assertPolynomValue(4, -4, 10, -3);

        assertPolynomValue(-1, 2, 1, 1);
    }

    @Test
    @Marking(points = 5)
    public void test_200_set_getValue() {

        assertPolynomValue(0, 0, -8, -10);
        assertPolynomValue(0, 0, -8, 100.);
        assertPolynomValue(0, 0, -8, -103.3);
        {
            final ObjectWrapper<SquarePolynom> poly = new ObjectWrapper<>(SquarePolynom.class, 0, 0, -8);
            poly.invoke(void.class, "setA", 7);
            poly.invoke(void.class, "setB", -1);
            final double expected =  2.5 * (7 * 2.5 - 1) - 8;
            final String msg = composeDescription(7, -1, -8).append(" should return " + expected).toString();
            Assert.assertEquals(msg, expected, poly.invoke(double.class, "get", 2.5), delta);
        }
    }

    @Test
    @Marking(points = 4)
    public void test_400_NoZero() {
        final ObjectWrapper<SquarePolynom> poly = new ObjectWrapper<>(SquarePolynom.class,1, 1, 1);
        assertNoZero(poly, composeDescription(1,2,1).toString());

        final ObjectWrapper<SquarePolynom> poly2 = new ObjectWrapper<>(SquarePolynom.class,1, 2, 3);
        assertNoZero(1, 6, 20, poly2);
        assertNoZero(2, 13, 41, poly2);
    }

    @Test
    @Marking(points = 1)
    public void test_500_OneZeroLinearOneZero() {
        final ObjectWrapper<SquarePolynom> poly = new ObjectWrapper<>(SquarePolynom.class,0, 1, 1);
        assertOneZero(poly, -1, composeDescription(0,1,1).toString());
    }

    @Test
    @Marking(points = 1)
    public void test_600_OneZeroLinearNoZero() {
        final ObjectWrapper<SquarePolynom> poly = new ObjectWrapper<>(SquarePolynom.class,0, 0, 1);
        assertNoZero(poly, composeDescription(0,0,1).toString());
    }

    @Test
    @Marking(points = 1)
    public void test_610_OneZero() {
        final ObjectWrapper<SquarePolynom> poly = new ObjectWrapper<>(SquarePolynom.class,1, 2, 1);
        assertOneZero(poly, -1, composeDescription(1,2,1).toString());

        final ObjectWrapper<SquarePolynom> poly2 = new ObjectWrapper<>(SquarePolynom.class,1, 2, 3);
        assertOneZero(6, -84, 294, poly2, 7);
        assertOneZero(21, 126, 189, poly2, -3);
    }

    @Test
    @Marking(points = 1)
    public void test_700_twoZeroes() {

        final ObjectWrapper<SquarePolynom> poly = new ObjectWrapper<>(SquarePolynom.class,1, 3, 2);
        assertTwoZeroes(poly, -2, -1, "x² + 3x + 2");

        assertTwoZeroes(3, 7, 4, poly, -4./3, -1);

        assertTwoZeroes(-2, 1, 3, poly, -1, 1.5);

        assertTwoZeroes(-1, 5, 6, poly, -1, 6);

        assertTwoZeroes(7, 11, 4, poly, -1, -4./7);

        assertTwoZeroes(4, -3, -10, poly, -1.25, 2);

        assertTwoZeroes(5, 6, 1, poly, -1, -0.2);

    }

    @Test
    @Marking(points = 1)
    public void test_800_Exception() {
        final ObjectWrapper<SquarePolynom> poly = new ObjectWrapper<>(SquarePolynom.class,0, 0, 0);

        final Class<ArithmeticException> expectedArithmeticException = ArithmeticException.class;

        // Either of a, b or c must be non-zero
        poly.invoke(double[].class, "getZeroes", expectedArithmeticException);
    }

    @Test
    @Marking(points = 1)
    public void test_900_TwoZeroExzess() {
        final ObjectWrapper<SquarePolynom> poly = new ObjectWrapper<>(SquarePolynom.class,
                46010, 598130,-1960026000);
        assertTwoZeroes(poly,-213,200,
                composeDescription(46010, 598130,-1960026000).toString());
    }

    // End of tests -------------------------------------------------------------------------------------------------
    // Test helper methods
    //
    static private final double delta = 1.E-12;
    private static void assertPolynomValue(int a, int b, int c, double x) {
        final ObjectWrapper<SquarePolynom> poly = new ObjectWrapper<>(SquarePolynom.class, a, b, c);
        final double expected = x * (a * x + b) + c;
        final String msg = composeDescription(a,b,c) + " for x == " + x + " should return " + expected;
        Assert.assertEquals(msg, expected, poly.invoke(double.class, "get", x), delta);
    }

    static private StringBuffer composeDescription(final int a, final int b, final int c) {

        final StringBuffer ret = new StringBuffer("p(x)=");

        boolean initialized = false;

        if (a != 0) {
            if (-1 == a) {
                ret.append("-");
            } else if (1 != a) {
                ret.append(a);
            }
            ret.append("x²");
            initialized = true;
        }

        if (0 != b) {
            if (initialized) {
                if (0 < b) {
                    ret.append("+");
                }
            } else {
                initialized = true;
            }
            if (-1 == b) {
                ret.append("-");
            } else if (1 != b) {
                ret.append(b);
            }
            ret.append("x");
        }

        if (0 != c) {
            if (initialized) {
                if (0 < c) {
                    ret.append("+");
                }
            }
            ret.append(c);
        }
        return ret;
    }

    /**
     * Testing given polynomial for no real zero value.
     *
     * @param poly Polynomial to be tested
     * @param description E.g. "-2x² + 4x -3"
     */
    static private void assertNoZero(final ObjectWrapper<SquarePolynom> poly, final String description) {

        Assert.assertEquals("No zero expected for polynom »" + description + "«",
                0, poly.invoke(double[].class, "getZeroes").length);
    }

    /**
     * Testing for no real valued zero solution.
     *
     * @param a Square coefficient
     * @param b Linear coefficient
     * @param c Constant coefficient
     * @param poly Square polynomial
     */
    static private void assertNoZero(final int a,
                                     final int b,
                                     final int c,
                                     final ObjectWrapper<SquarePolynom> poly) {
        poly.invoke(void.class, "setA", a);
        poly.invoke(void.class, "setB", b);
        poly.invoke(void.class, "setC", c);
        Assert.assertEquals("One zero expected for polynom »" + composeDescription(a, b, c) + "«",
                0, poly.invoke(double[].class, "getZeroes").length);


    }

    /**
     * Testing given polynomial for one zeroes to be returned
     *
     * @param poly Polynomial to be tested
     * @param expected zero value
     * @param description E.g. "-2x² + 4x -3"
     */
    static private void assertOneZero(final ObjectWrapper<SquarePolynom> poly,
                                      double expected,
                                      final String description) {
        final double[] result = poly.invoke(double[].class, "getZeroes");

        Assert.assertEquals("One zero expected for polynom »" + description + "«",
                1, result.length);
        Assert.assertEquals("Expected zero of »" + description +"« to be " + expected, expected, result[0], delta);
    }

    /**
     * Like {@link #assertOneZero(ObjectWrapper, double, String)} but re-setting coefficients beforehand,
     * @param a Square coefficient
     * @param b Linear coefficient
     * @param c Constant coefficient
     * @param poly Square polynomial
     */
    static private void assertOneZero(final int a,
                                      final int b,
                                      final int c,
                                      final ObjectWrapper<SquarePolynom> poly,
                                      double expected) {
        poly.invoke(void.class, "setA", a);
        poly.invoke(void.class, "setB", b);
        poly.invoke(void.class, "setC", c);
        final double[] result = poly.invoke(double[].class, "getZeroes");

        final StringBuffer polynomDescription = composeDescription(a, b, c);

        Assert.assertEquals(2 + " zeroes expected for polynom »" + polynomDescription + "«",
                1, result.length);
        Assert.assertEquals("Expected zero of »" + polynomDescription +"« to be " + expected,
                expected, result[0], delta);
    }

    /**
     * Testing given polynomial for two zeroes to be returned
     *
     * @param poly Polynomial to be tested
     * @param expectedLower Lower zero value
     * @param expectedUpper Upper zero value
     * @param description E.g. "-2x² + 4x -3"
     */
    static private void assertTwoZeroes(final ObjectWrapper<SquarePolynom> poly,
                                        double expectedLower,
                                        double expectedUpper,
                                        final String description) {
        final double[] result = poly.invoke(double[].class, "getZeroes");

        Assert.assertEquals(2 + " zeroes expected for polynom »" + description + "«",
                2, result.length);
        Assert.assertEquals("Expected lower zero of »" + description +"« to be " + expectedLower,
                expectedLower, result[0], delta);
        Assert.assertEquals("Expected upper zero of »" + description +"« to be " + expectedUpper,
                expectedUpper, result[1], delta);
    }

    /**
     * Like {@link #assertTwoZeroes(ObjectWrapper, double, double, String)} but re-setting coeficients beforehand,
     * @param a Square coefficient
     * @param b Linear coefficient
     * @param c Constant coefficient
     * @param poly Square polynomial
     * @param expectedLower Expected lower zero value
     * @param expectedUpper Expected upper zero value
     */
    static private void assertTwoZeroes(final int a,
                                        final int b,
                                        final int c,
                                        final ObjectWrapper<SquarePolynom> poly,
                                        double expectedLower,
                                        double expectedUpper) {
        poly.invoke(void.class, "setA", a);
        poly.invoke(void.class, "setB", b);
        poly.invoke(void.class, "setC", c);
        final double[] result = poly.invoke(double[].class, "getZeroes");

        final StringBuffer polynomDescription = composeDescription(a, b, c);

        Assert.assertEquals(2 + " zeroes expected for polynom »" + polynomDescription + "«",
                2, result.length);
        Assert.assertEquals("Expected lower zero of »" + polynomDescription +"« to be " + expectedLower,
                expectedLower, result[0], delta);
        Assert.assertEquals("Expected upper zero of »" + polynomDescription +"« to be " + expectedUpper,
                expectedUpper, result[1], delta);
    }
}
